/*
 * Copyright 2013-2018 Radian Memory Systems, Inc.  All Rights Reserved.
 *
 * This file is provided under non-disclosure agreement.
 * See the file COPYING for details.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <unistd.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <asm-generic/types.h>
typedef __u64 u64;
typedef __u32 u32;
typedef __u16 u16;
typedef __u8 u8;
#include "nvme_express.h"
#include "nvme_ioctl.h"
#include "dialogDef.h"
#include "dialog_util.h"
#define MAX_DIES 256
#include "RMSVendorSpecific.h"
#include "nvme_util.h"
#include "rms200_hw_info.h"

#define XFER_BUF (4 * 1024)    /* Size of debug message buffer */
#define CONTROL_DEV		"/dev/nvme0"
#define MAX_BOARD_NAME		256
#define MAGIC 0x43212143

static int cmdId = 1;

static void
wait_for_q(void)
{
	system("/bin/stty raw");
	getchar();
	system("/bin/stty cooked");
}

static void
output_fw_slot_info(struct firmware_slot_log *x)
{
	printf("activeSlotNumber = %d\n", x->activeFirmwareInfo & 0x7);
	if (x->activeFirmwareInfo & 0x70) {
		printf("nextActiveSlotNumber = %d\n",
				x->activeFirmwareInfo >> 4);
	}
	printf("slot1Version = %.8s\n", x->FirmwareRevisionSlot1);
	printf("slot2Version = %.8s\n", x->FirmwareRevisionSlot2);
	printf("slot3Version = %.8s\n", x->FirmwareRevisionSlot3);
	printf("slot4Version = %.8s\n", x->FirmwareRevisionSlot4);
	printf("Slots 5-7 unsupported\n");
}

static void
output_128(char *str, unsigned char *val)
{
	unsigned long long upper = *(unsigned long long *)(val + 8);
	unsigned long long lower = *(unsigned long long *)(val);

	if (upper == 0)
		printf("%s%lld\n", str, lower);
	else
		printf("%s0x%llx %.16llx\n", str, upper, lower);
}

static void
output_smart_log(struct smart_log *x)
{
	printf("criticalErrorVector = 0x%x\n", x->criticalError);
	printf("temperature = %d Kelvin\n", x->temperature);
	printf("availableSpace = %d%%\n", x->availableSpace);
	printf("availableSpaceThreshold = %d%%\n",
			x->availableSpaceThreshold);
	output_128("numberOf512ByteBlocksRead = ", x->dataUnitsRead);
	output_128("numberOf512ByteBlocksWritten = ",
			x->dataUnitsWritten);
	output_128("numberOfHostReadCommands = ", x->hostReadCommands);
	output_128("numberOfHostWriteCommands = ", x->hostWriteCommands);
	output_128("controllerBusyTimeMinutes = ", x->controllerBusyTime);
	output_128("numberOfPowerCycles = ", x->powerCycles);
	output_128("powerOnHours = ", x->powerOnHours);
	output_128("unsafeShutdowns = ", x->unsafeShutdowns);
	output_128("mediaErrors = ", x->mediaErrors);
	output_128("numberOfErrorInfoLogs = ", x->numberOfErrorInfoLogs);
}

static void
output_error_log(struct error_log *x)
{
	int i;
	for (i = 0; i < 63; i++) {
		if (x->errorCount == 0)
			break;
		printf("errorCount = %lld\n", x->errorCount);
		printf("sqID = 0x%x\n", x->sqID);
		printf("cmdID = 0x%x\n", x->cmdID);
		printf("status = 0x%x\n", x->status);
		printf("errorByte = 0x%x\n", x->errorByte);
		printf("errorBit = 0x%x\n", x->errorBit);
		printf("lba = %lld\n", x->lba);
		printf("namespace = %d\n", x->nameSpace);
		x++;
	}
	printf("numOfErrorLogEntries = %d\n", i);
}

static void
create_get_log_page_cmd(
	struct usr_io *x,
	unsigned char *buf,
	unsigned long len,
	unsigned long id,
	unsigned long word12)
{
	memset(x, 0x0, sizeof(struct usr_io));
	x->cmd.header.opCode = NVM_ADMIN_CMD_GET_LOG_PAGE;
	x->cmd.header.cmdID = cmdId++;
	if (id == GLP_ID_SMART_HEALTH)
		x->cmd.header.namespaceID = 0xffffffff; /* Global ns ID */
	x->direction = XFER_FROM_DEV;
	x->timeout = 1;
	x->length = len;
	x->addr = (__u64) buf;
	x->cmd.cmd.getLogPage.numDW = (len >> 2) - 1;
	x->cmd.cmd.getLogPage.LogPageID = id;
	x->cmd.dw[12] = word12;
}

static void
create_rms200_get_hw_info_cmd(
	struct usr_io *x,
	unsigned char *buf,
	unsigned long len)
{
	nvme_cmd_rms200_t nvme_cmd_rms200;

	create_generic_uio_pkt(x, XFER_FROM_DEV, buf, len);

	memset(&nvme_cmd_rms200, 0x0, sizeof(nvme_cmd_rms200));
	nvme_cmd_rms200.subOpcode = RMS_GET_HW_INFO;
	nvme_cmd_rms200.numOfDwordPrP = (len >> 2) - 1;

	append_rms_cmd_pkt(x, &nvme_cmd_rms200);
}

#ifndef DIALOG_V3
static int
output_snapshot_trailer(
	int fd,
	DialogPage *dp)
{
	struct usr_io x;
	struct rms200_hw_info hw_info;
	int status;
	static char modnum[9];
	static char sernum[9];
	char fw_rev[9];
	char hw_rev[9];
	static int first_time = 1;

	if (first_time) {
		modnum[8] = sernum[8] =  '\0';

		first_time--;

		/* Model number and serial number never change */

		create_rms200_get_hw_info_cmd(&x,
				(unsigned char *) &hw_info,
				sizeof(struct rms200_hw_info));
		status = ioctl(fd, NVME_IOCTL_ADMIN_CMD, &x);
		if (status != 0)
			return status;
		if (x.comp.SCT != 0 || x.comp.SC != 0)
			return status;

		sprintf(modnum, "%x", hw_info.model_number);
		memcpy(sernum, hw_info.serial_number, 8);
	}
	fw_rev[8] = hw_rev[8] =  '\0';
	memcpy(fw_rev, dp->dpad.extensions.firmwareRev, 8);
	memcpy(hw_rev, dp->dpad.extensions.hardwareRev, 8);
	printf("-----------------------\n");
	printf("modelNumber = RMS-%s\n", modnum);
	printf("serialNumber = %.8s\n", sernum);
	printf("hardwareRevision = %.8s\n", hw_rev);
	printf("firmwareVersion = %.8s\n", fw_rev);
	return 0;
}
#endif

static void
get_board_trailer(
	int fd,
	char *trailer)
{
	struct usr_io x;
	struct rms200_hw_info hw_info;
	struct firmware_slot_log fw_log;
	char   fw_version[32];
	int    fw_slot;
	int status;

	trailer[0] = 0;

	create_rms200_get_hw_info_cmd(&x,
			(unsigned char *) &hw_info,
			sizeof(struct rms200_hw_info));
	status = ioctl(fd, NVME_IOCTL_ADMIN_CMD, &x);
	if (status != 0)
		return;
	if (x.comp.SCT != 0 || x.comp.SC != 0)
		return;

	create_get_log_page_cmd(&x,
			(unsigned char *) &fw_log,
			sizeof(struct firmware_slot_log),
			GLP_ID_FIRMWARE_SLOT_INFO,
			0);
	status = ioctl(fd, NVME_IOCTL_ADMIN_CMD, &x);
	if (status != 0)
		return;
	if (x.comp.SCT != 0 || x.comp.SC != 0)
		return;

	fw_slot = fw_log.activeFirmwareInfo & 0x7;
	switch (fw_slot) {
	case 1:
		sprintf(fw_version, "%.8s", fw_log.FirmwareRevisionSlot1);
		break;
	case 2:
		sprintf(fw_version, "%.8s", fw_log.FirmwareRevisionSlot2);
		break;
	case 3:
		sprintf(fw_version, "%.8s", fw_log.FirmwareRevisionSlot3);
		break;
	case 4:
		sprintf(fw_version, "%.8s", fw_log.FirmwareRevisionSlot4);
		break;
	default:
		sprintf(fw_version, "????????");
		break;
	}

	sprintf(trailer, "-----------------------\n");
	trailer += strlen(trailer);
	sprintf(trailer, "modelNumber = RMS-%x\n",
			hw_info.model_number);
	trailer += strlen(trailer);
	sprintf(trailer, "serialNumber = %.8s\n",
			hw_info.serial_number);
	trailer += strlen(trailer);
	sprintf(trailer, "hardwareRevision = %.8s\n",
			hw_info.top_rev_number);
	trailer += strlen(trailer);
	sprintf(trailer, "firmwareVersion = %.8s on slot %d\n",
			fw_version, fw_slot);
}

static inline int isICRvalidinPage(DialogPage *dp)
{
    return (dp->initialCapacitance > 0) && (dp->initialEsr > 0);
}

static void
description(void)
{
	printf("Dump log page from device\n");
	return;
}

static void
usage(void)
{
	printf("usage:  dump_log_page [devName] -s 1-3 | -d 0-127 [-x]\n");
	printf("   devName: The device to use - default /dev/nvme0\n");
	printf("   -d <n> : Dialog page. 0 for current, >0 for snapshots\n");
	printf("   -s <n> : standard nvme log page <n>, where <n> is 1-3\n");
	printf("      -s 1: dump error info\n");
	printf("      -s 2: dump smart health info\n");
	printf("      -s 3: firmware slot info\n");
	printf("   -x     : dump dialog in XML format\n");
	printf("   -f     : dialog.xml is used as the XML name\n");
	printf("   -t     : dump only temperatures\n");
	printf("   -p     : dump only power on hours\n");
	printf("   -a     : dump all dialog snapshots\n");
	printf("   -l     : list dialog snapshots\n");
	printf("   -i     : dump only identification information\n");
	printf("   Example: dump_log_page -d 0\n");
}

int
main(int argc, char **argv)
{
	struct usr_io x;
	unsigned char *buffer = (unsigned char *) 0;
	int fd = 0;
	int status;
	int id;
	int buflen;
	int c;
	struct firmware_slot_log *firmware_slot_log_buf;
	struct smart_log *smart_log_buf;
	struct error_log *error_log_buf;
	DialogPage *dp;
	int nvme_std_log = 0;
	int dialog_in_xml = 0;
	int dialog_xml_file = 0;
	char *control_dev = CONTROL_DEV;
	int word12 = 0;
	char *pEnv;
	int  all_pages = 0;
	int  count = 1;
#ifndef DIALOG_V3
	int is_snapshot = 0;
	u32 last_snaptime = 0;
	u32 last_nexttime = 0;
	int  date_only = 0;
	int  id_only = 0;
	int  temperature_only = 0;
	int  power_on_only = 0;
#endif
	char trailer[256];
	int i;

	/* Override default device */
	if (argc >= 2 && strncmp("/dev/", argv[1], 5) == 0) {
		control_dev = argv[1];
	} else {
		pEnv = getenv("RMS_DEVICE");
		if (pEnv) {
			control_dev = pEnv;
		}
	}

	while (1) {

		c = getopt(argc, argv, "ixd:fs:tphHal");

		if (c == -1)
			break;

		switch (c) {

		case 'd':
			id = GLP_ID_DIALOG;
			nvme_std_log = 1;
			word12 = atoi(optarg);
#ifndef DIALOG_V3
			if (word12)
				is_snapshot = 1;
#endif
			if (word12 > 255 || word12 < 0) {
				usage();
				exit(1);
			}
			break;

		case 'f':
			dialog_xml_file = 1;
			break;

		case 's':
			id = atoi(optarg);
			nvme_std_log = 1;
			if (id > 3 || id < 1) {
				usage();
				exit(1);
			}
			break;

		case 'x':
			dialog_in_xml = 1;
			break;

		case 'a':
			id = GLP_ID_DIALOG;
			nvme_std_log = 1;
			all_pages = 1;
#ifndef DIALOG_V3
			is_snapshot = 1;
#endif
			break;

		case 'l':
			id = GLP_ID_DIALOG;
			nvme_std_log = 1;
			all_pages = 1;
#ifndef DIALOG_V3
			date_only = 1;
#endif
			break;

#ifndef DIALOG_V3
		case 't':
			temperature_only = 1;
			break;

		case 'p':
			power_on_only = 1;
			break;

		case 'i':
			id_only = 1;
			break;
#endif
		case 'H':
			description();
			usage();
			wait_for_q();
			exit(1);
			break;

		case 'h':
		default:
			usage();
			exit(1);
		}
	}


	if (nvme_std_log == 0) {
		usage();
		exit(1);
	}

	if (posix_memalign((void **) &buffer, XFER_BUF, XFER_BUF) != 0) {
		printf("Memory allocation error\n");
		exit(1);
	}

	firmware_slot_log_buf = (struct firmware_slot_log *) buffer;
	smart_log_buf = (struct smart_log *) buffer;
	error_log_buf = (struct error_log *) buffer;
	dp = (DialogPage *) buffer;

	fd = open(control_dev, O_RDWR);
	if (fd < 0) {
		printf("Could not open file %s ", control_dev);
		if (errno == EACCES)
			printf(" Access error - try running as su\n");
		else
			printf(" Error %d\n", errno);
		exit(1);
	}

	switch (id) {
	case GLP_ID_FIRMWARE_SLOT_INFO:
		buflen = sizeof(struct firmware_slot_log);
		break;

	case GLP_ID_SMART_HEALTH:
		buflen = sizeof(struct smart_log);
		break;

	case GLP_ID_ERR_INFO:
		/* error log struct is 64 bytes for 1 entry */
		buflen = sizeof(struct error_log);
		/* IDT supports 63 entries */
		buflen *= 63;
		break;

	case GLP_ID_DIALOG:
		/* Dialog log struct is less than 4k bytes */
		buflen = sizeof(DialogPage);
		break;
	}

	get_board_trailer(fd, trailer);

	do {
		/* Dump a standard nvme log pages */
		if (all_pages)
			word12 = count;
		create_get_log_page_cmd(&x, buffer, buflen, id, word12);
		status = ioctl(fd, NVME_IOCTL_ADMIN_CMD, &x);
		if (status != 0) {
			printf("status = %d\n", status);
			goto done;
		}

		if (x.comp.SCT != 0 || x.comp.SC != 0) {
			if (!all_pages) {
				printf("operationFailedNVMeError = %s\n",
					nvme_err(x.comp.SCT, x.comp.SC));
			}
			goto done;
		}

		switch (id) {
		case GLP_ID_FIRMWARE_SLOT_INFO:
			output_fw_slot_info(firmware_slot_log_buf);
			break;

		case GLP_ID_SMART_HEALTH:
			output_smart_log(smart_log_buf);
			break;

		case GLP_ID_ERR_INFO:
			output_error_log(error_log_buf);
			break;

		case GLP_ID_DIALOG:
			if (dialog_in_xml) {
				output_dialog_xml(dp,
						dialog_xml_file
						? NULL
						: stdout);
			}
#ifndef DIALOG_V3
			else if (temperature_only) {
				output_dialog_temperatures(dp, stdout);
			} else if (power_on_only) {
				output_dialog_power_on(dp, stdout);
			} else if (date_only) {
				/*
				 * Old versions of firmware don't
				 * return error if snapshot is out of
				 * range.  This checks to see if this
				 * snapshot is actually the previous.
				 */
				if (dp->snapshotTime != last_snaptime
				 || dp->nextSnapshotTime != last_nexttime) {
					last_snaptime = dp->snapshotTime;
					last_nexttime = dp->nextSnapshotTime;
					output_dialog_date(dp, count, stdout);
				}
			} else if (id_only) {
				fprintf(stdout, trailer);
			}
#endif

			else {
				if (all_pages && count > 1) {
					/*
					 * Put some vertical space between
					 * log pages
					 */
					printf("\n\n");
					for (i = 0; i < 79; ++i)
						putchar('=');
					printf("\n\n\n");
				}
				output_dialog(dp, stdout);
#ifndef DIALOG_V3
				if (is_snapshot
				 && dp->dpad.extensions.nDwords) {
					output_snapshot_trailer(fd, dp);
				} else
#endif
					fprintf(stdout, "%s", trailer);
			}
			break;
		}
		count++;
	} while (all_pages);

done:
	if (fd > 0)
		close(fd);
	if (buffer)
		free(buffer);
	exit(0);
}
